import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

sns.set(style="darkgrid")
"""
Function to plot the loss and MAPE

=============
Input : list for losses and arguments

Ouptut :
=============

"""
def plot(train_loss_list, train_mape_t_list, train_mape_n_list, dev_loss_list, dev_mape_t_list, dev_mape_n_list):
    f1 = plt.figure(1)
    plt.plot(train_loss_list)
    plt.plot(dev_loss_list)
    plt.title('Loss')
    plt.ylabel('Loss')
    plt.xlabel('epoch')
    plt.legend(['Train', 'Dev'], loc='upper left')
    # plt.show()
    # plt.close(f1)

    f2 = plt.figure(2)
    plt.plot(train_mape_t_list)
    plt.plot(dev_mape_t_list)
    plt.title('MAPE_t')
    plt.ylabel('MAPE_t')
    plt.xlabel('epoch')
    plt.legend(['Train', 'Dev'], loc='upper left')
    # plt.show()
    # plt.close(f2)

    f3 = plt.figure(3)
    plt.plot(train_mape_n_list)
    plt.plot(dev_mape_n_list)
    plt.title('MAPE_n')
    plt.ylabel('MAPE_n')
    plt.xlabel('epoch')
    plt.legend(['Train', 'Dev'], loc='upper left')
    plt.show()
    plt.close(f1)
    plt.close(f2)
    plt.close(f3)

def plot_host_events(events, times):
    host_cnt = len(events)
    plt.figure()
    for i,(event,time) in enumerate(zip(events,times)):
        plt.subplot(host_cnt,1,i+1)
        sns.lineplot(time,event)
        plt.xlabel("Time")
        plt.ylabel("Length of Packets")
        plt.title(f"Time series data for Host {i}")
    plt.show()

def plot_predictions(data):
    temporal_prediction = np.transpose(data,axes=(1,0))
    plt.figure()
    host_cnt = len(temporal_prediction)
    time_steps = len(temporal_prediction[0])
    for i,time in enumerate(temporal_prediction):
        plt.subplot(host_cnt,1,i+1)
        sns.lineplot(range(time_steps), time,marker='o')
        plt.xlabel("Time stamp")
        plt.ylabel("Time of Events")
        plt.title(f"Time predictions for Host {i}")
    plt.show()

if __name__ == "__main__":
    x_t = [1,2,3,4]
    x_d = [2,3,4,5]
    y_t = [4,3,2,1]
    y_d = [3,2,1,0]
    z_t = [1,-1,2,-2]
    z_d = [-1,2,-2,3]    
    plot(x_t,y_t,z_t,x_d,y_d,z_d)


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
"""
Class of model used -- NTPP

"""


class NTPP(nn.Module):

    def __init__(self, args, output_layer_size=1):
        super(NTPP,self).__init__()
        self.hidden_size = args['h']
        self.num_layers = args['nl']
        self.lstm_step = nn.LSTM(args['element_size'], args['h'], args['nl'], batch_first=True)
        self.fc = nn.Linear(args['h'], output_layer_size) # output_layer_size = number of host

    #  forward pass
    def forward(self,x):

        # Set initial hidden and cell states
        h0 = torch.zeros(self.num_layers, x.size(0), self.hidden_size)
        c0 = torch.zeros(self.num_layers, x.size(0), self.hidden_size)

        # Forward propagate LSTM
        out, _ = self.lstm_step(x, (h0, c0))  # out: tensor of shape (batch_size, seq_length, hidden_size)

        # Decode the hidden state of the last time step
        out = self.fc(out)
        out = torch.exp(out.squeeze())
        return out


import argparse, time
import torch
import numpy as np
import random
from augurnet.data import NTPPData
from augurnet.utils import ensure_dir
from augurnet.data import TemporalData ######
from augurnet.model import NTPP
from augurnet.scorer import discriminatorLoss, calculateLoss
from augurnet.plotter import plot, plot_predictions

class Augur:
    args = {
        'save_dir':'augurnet/data/saved/',
        'window_size':64,
        'int_count':100,
        'test_size':0.2,
        'time_step':8,
        'batch_size':11,
        'element_size':2,
        'h':32,
        'nl':1,
        'seed':123456,
        # 'mode',
        'epochs':100,
        'workers':4,
        'learning_rate':0.01,
        'metric':'AUC',
        'is_cuda':True,
        'optim':'Adam',
    }
    def __init__(self, **kwrags):
        # init. the paramters and the torch environment
        self.update(kwrags)
        torch.set_default_tensor_type('torch.DoubleTensor')

        np.random.seed(Augur.args['seed'])
        torch.manual_seed(Augur.args['seed'])
        random.seed(Augur.args['seed'])
        self.device = torch.device('cuda') if torch.cuda.is_available() and Augur.args['is_cuda'] else torch.device('cpu')
    
    def update(self,args):
        for arg,value in args.items():
            if arg in Augur.args:
                Augur.args[arg] = value

    def print_args(self):
        print(Augur.args)
    
    def fit(self,temporal_data,verbose=0, save_file=None):
        #  Code for training the model
        #model file --
        self.data = NTPPData(temporal_data, Augur.args)
        ensure_dir(Augur.args['save_dir'])
        #Loading Data

        #  Get the obsevation
        train_y,test_y = self.data.getObservation()
        if verbose: print('NTPP model...')
        self.model = NTPP(Augur.args, output_layer_size=1)

        #  Data using dataloader
        train_loader = torch.utils.data.DataLoader(self.data,
                                batch_size=Augur.args['batch_size'],
                                shuffle=False,
                                num_workers=Augur.args['workers'],
                                pin_memory=Augur.args['is_cuda'] # CUDA only
                                )
        #  Depending on the argument decide the optimizer
        if Augur.args['optim'] == 'Adam':
            optimizer = torch.optim.Adam(
                self.model.parameters(), lr=Augur.args['learning_rate'])
        elif Augur.args['optim'] == 'SGD':
            optimizer = torch.optim.SGD(
                self.model.parameters(), lr=Augur.args['learning_rate'])
        elif Augur.args['optim'] == 'RMS':
            optimizer = torch.optim.RMSprop(
                self.model.parameters(), lr=Augur.args['learning_rate'])
        else:
            raise Exception('use Proper optimizer')
        # Data struture to store teh loss and MAPE values of Dev and train set
        train_loss_list,train_mape_t_list, train_mape_n_list = [],[],[]
        dev_loss_list,dev_mape_t_list, dev_mape_n_list = [],[],[]

        # For each epochs
        for epoch_cnt,epoch in enumerate(range(Augur.args['epochs'])):
            #  Set the dataloader to Train
            self.data.startTrain()
            for i, (batch_events,batch_times) in enumerate(train_loader):
                start_time = time.time()
                # batch preprocess
                time_step = Augur.args['time_step']

                # Some preprocessing
                batch_events_part1 = batch_events[:,:time_step]
                batch_times_diff = batch_times[:, 1:1+time_step] - batch_times[:,:time_step]
                batch_times_diff_next = batch_times[:, 2:2+time_step] - batch_times[:,1:1+time_step]
                batch_times_diff = batch_times_diff[:,:,None] # exapnd dim in axis 2
                batch_events_part1 = batch_events_part1[:,:,None] # expand Dim
                batch_input = torch.cat((batch_times_diff,batch_events_part1),2)

                #forward pass
                outputs = self.model(batch_input)
                # print("outputs",outputs.size())
                last_time_step_layer = outputs.clone()
                last_time_step_layer = (last_time_step_layer.detach().numpy()).transpose()[-1]
                # print("Last_time_step",[1/(x+1e-9) for x in last_time_step_layer])
                predicted = []
                # print("last_time_step_layer shape: ", len(last_time_step_layer))
                for host in range(last_time_step_layer.shape[0]):
                    for secon_host in range(host+1,last_time_step_layer.shape[0]):
                        predicted.append(np.greater(last_time_step_layer[host],last_time_step_layer[secon_host]))
                discriminator_loss = discriminatorLoss(train_y, predicted, Augur.args['metric'])
                loss, mape_t, mape_n = calculateLoss(discriminator_loss, outputs, batch_times_diff_next, time_step)

                # save the loss for the plotting
                train_loss_list.append(loss)
                train_mape_t_list.append(mape_t)
                train_mape_n_list.append(mape_n)

                #backward and optimize
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                if epoch_cnt%100==0:
                    if verbose:  print('[TRAIN] Epoch [{}/{}], Step [{}/{}], Loss: {:.4f} MAPE_t: {:4f} MAPE_n: {:4f} Time : {}'
                        .format(epoch, Augur.args['epochs'], i+1, len(train_loader), loss.item(), mape_t.item(), mape_n.item(), time.time()-start_time))

            # Set the dataoader to dev set
            self.data.startDev()
            with torch.no_grad():
                for i, (batch_events,batch_times) in enumerate(train_loader):
                    start_time = time.time()
                    # batch preprocess
                    time_step = Augur.args['time_step']
                    batch_events_part1 = batch_events[:,:time_step]
                    batch_times_diff = batch_times[:, 1:1+time_step] - batch_times[:,:time_step]
                    batch_times_diff_next = batch_times[:, 2:2+time_step] - batch_times[:,1:1+time_step]
                    batch_times_diff = batch_times_diff[:,:,None] # exapnd dim in axis 2
                    batch_events_part1 = batch_events_part1[:,:,None] # expand Dim
                    batch_input = torch.cat((batch_times_diff,batch_events_part1),2)

                    #forward pass
                    outputs = self.model(batch_input)
                    last_time_step_layer = outputs.clone()
                    last_time_step_layer = (last_time_step_layer.detach().numpy()).transpose()[-1]
                    predicted = []
                    for host in range(last_time_step_layer.shape[0]):
                        for secon_host in range(host+1,last_time_step_layer.shape[0]):
                            predicted.append(np.greater(last_time_step_layer[host],last_time_step_layer[secon_host]))
                    discriminator_loss = discriminatorLoss(train_y, predicted, Augur.args['metric'])
                    loss, mape_t, mape_n = calculateLoss(discriminator_loss, outputs, batch_times_diff_next, time_step)
                    dev_loss_list.append(loss)
                    dev_mape_t_list.append(mape_t)
                    dev_mape_n_list.append(mape_n)
                    if epoch_cnt%100==0:
                        if verbose:  print('[VALIDATION] Epoch [{}/{}], Step [{}/{}], Loss: {:.4f} MAPE_t: {:4f} MAPE_n: {:4f} Time : {}'
                            .format(epoch, Augur.args['epochs'], i+1, len(train_loader), loss.item(), mape_t.item(), mape_n.item(), time.time()-start_time))

        # Plot the loss and MAPE
        self.train_meta_data = [train_loss_list, train_mape_t_list, train_mape_n_list,
            dev_loss_list, dev_mape_t_list, dev_mape_n_list]
        if save_file is not None:
            torch.save(self.model, save_file)

    def plot_losses(self):
        plot(*self.train_meta_data)

    def predict(self, time_step=10, model_file=None, verbose=0):
        if model_file is not None:
            self.model = torch.load(model_file)
        
    #  Data using dataloader
        train_loader = torch.utils.data.DataLoader(self.data,
                                batch_size=Augur.args['batch_size'],
                                shuffle=False,
                                num_workers=Augur.args['workers'],
                                pin_memory=Augur.args['is_cuda'] # CUDA only
                                )
        
        self.data.startTrain()
        self.time_predictions = []
        for run in range(time_step):
            for i, (batch_events,batch_times) in enumerate(train_loader):
                start_time = time.time()
                # batch preprocess 
                time_step = Augur.args['time_step']

                # Some preprocessing
                batch_events_part1 = batch_events[:,:time_step]
                batch_times_diff = batch_times[:, 1:1+time_step] - batch_times[:,:time_step]
                batch_times_diff = batch_times_diff[:,:,None] # exapnd dim in axis 2
                batch_events_part1 = batch_events_part1[:,:,None] # expand Dim
                batch_input = torch.cat((batch_times_diff,batch_events_part1),2)

                #forward pass
                outputs = self.model(batch_input)
                # print("outputs",outputs.size())
                last_time_step_layer = outputs.clone()
                last_time_step_layer = (last_time_step_layer.detach().numpy()).transpose()[-1]
                times = [1/(x+1e-9) for x in last_time_step_layer]
                if verbose: print("Last_time_step",times)
                self.time_predictions.append(times)
                self.data.change_data(times)

    def plot_next_events(self,*args,**kwrags):
        plot_predictions(self.time_predictions)


if __name__ == "__main__":
    events_filename = "augur/data/preprocess/events.txt"
    times_filename  = "augur/data/preprocess/times.txt"
    data = TemporalData(events_filename, times_filename)
    print("Host count",data.get_host_count())
    data.plot_data()
    model = Augur()
    model.fit(data)
    model.plot_losses()
    model.predict(verbose=1)
    model.plot_next_events()
            


from sklearn.metrics import roc_auc_score, precision_score, recall_score
import torch
import numpy as np
import sys

"""
Function to find the discriminator loss

=============
Input : training input, predicted value, metric used
output : loss (depending on metric)

Ouptut :
=============

"""
def discriminatorLoss(train_y, predicted, metrics='AUC'):
    res = []
    for i in range(len(train_y)):
        if metrics == 'AUC':
            # try:
            # print("Train_y: ",train_y[i])
            # print("predicted: ",predicted)
            # sys.exit()
            res.append(roc_auc_score(train_y[i], predicted))
            # except:
            # pass# print("Exception")
        elif metrics == 'PRECISION':
            res.append(precision_score(train_y[i], predicted))
        elif metrics == 'RECALL':
            res.append(recall_score(train_y[i], predicted))
        else:
            raise Exception('NotACorrectMetrics')
    return res


"""
Function to calculate the loss

=============
Input : discriminator loss, output from the model, difference between events, time step

Ouptut : Loss, MAPE_t, MAPE_n
=============

"""
def calculateLoss(dLoss, output, batch_times_diff_next, time_step):
    output_delta_mul = torch.mul(output, batch_times_diff_next)
    # print(output.size(),batch_times_diff_next.size())
    # print("output_delta_mul",output_delta_mul)
    time_LLs = (torch.log(output) - output_delta_mul)
    Nhhat = output_delta_mul.sum(dim=1)
    batch_size = output.shape[0]
    Nh = torch.tensor([time_step * 1.0] * batch_size)
    mape_n = (torch.abs(Nh - Nhhat) / Nh).mean()

    pred_delta = torch.abs(1 / output).mean()
    obs_delta = torch.abs(batch_times_diff_next).mean()
    mape_t = torch.abs(pred_delta - obs_delta) / obs_delta
    # print("output {0}\noutput{1}".format(torch.log(output), output))
    loss = ((-time_LLs) + (torch.tensor(dLoss)).mean()).mean()
    # sys.exit()
    return loss, mape_t, mape_n



import pandas as pd
import os
import numpy as np
import subprocess
from collections import defaultdict

"""
Function to form events.txt and times.txt from .csv input file

=============
Input : filename and threshold

Ouptut :
=============

"""


def extractor(path, minl=50):
    data = pd.read_csv(path + '/testpackets.csv')
    event_time_stamps = {}
    for row in range(data.shape[0]):
        row = list(data.loc[row])
        if row[2] in event_time_stamps:
            event_time_stamps[row[2]].append((row[5], row[1]))
        else:
            event_time_stamps[row[2]] = [(row[5], row[1])]
    filtered_event_time_stamps = {
        k: v
        for k, v in event_time_stamps.items() if len(v) >= minl
    }
    with open(path + '/ntpp/preprocess/events.txt', 'w') as f:
        for k in filtered_event_time_stamps.keys():
            f.write(" ".join(
                [str(x[0]) for x in filtered_event_time_stamps[k]]))
            f.write("\n")

    with open(path + '/ntpp/preprocess/times.txt', 'w') as f:
        for k in filtered_event_time_stamps.keys():
            f.write(" ".join(
                [str(x[1]) for x in filtered_event_time_stamps[k]]))
            f.write("\n")


"""
Function to read the events.txt and times.txt and load then in Data struture

=============
Input : filename

Ouptut : list of list
=============

"""
def read_file(filename):
    rows = open(filename,'r').readlines()
    data = [[float(ti) for ti in line.strip().split(' ')] for line in rows]
    return np.array(data)


"""
Function to comapre the event value of host with rest of host

=============
Input : indexes of times, number of host, data

Ouptut : list of list
=============

"""
def compare_interval_count(left, right, host_count, interval_count):
    Y = []
    for interval in range(left, right):
        a = []
        for host in range(host_count - 1):
            a.extend(
                np.greater(interval_count[host,interval],
                           interval_count[:, interval])[host + 1:])
        # print("a :", len(a))
        Y.append(a)
    return Y


"""
Function to ensure of the directory exist or not

=============
Input : directory

Ouptut :
=============

"""
def ensure_dir(d, verbose=True):
    if not os.path.exists(d):
        if verbose:
            print("Directory {} do not exist; creating...".format(d))

        os.makedirs(d)

def getMinCount(filename):
    events = read_file(filename)
    min_count = min([len(x) for x in events])
    return min_count

def read_pcap(filename):
    cmd = f"""tshark -r {filename} -E separator=, -E occurrence=f -E quote=d -o 'gui.column.format:"Time","%t","Source","%s", "Length","%L"'"""
    res = subprocess.check_output(cmd,shell=True).decode('utf-8')
    res = res.strip().split("\n")
    events, times = defaultdict(list),defaultdict(list)
    for entry in res:
        print(entry)
        if len(entry)>0:
            entry = entry.strip()
            [time, source, length] = entry.split(" ")
            events[source].append(int(length))
            times[source].append(float(time))
    return (list(events.values())),(list(times.values()))
    
if __name__ =="__main__":
    filename = "augur/data/fuzz-2006-11-24-4742.pcap"
    print(read_pcap(filename))

    



import numpy as np 
from augurnet.utils import read_file, compare_interval_count
from augurnet.plotter import plot_host_events
from os import path
import errno
import os
import logging
import torch
from torch.utils.data.dataset import Dataset


class TemporalData:
    def __init__(self, events_filename, times_filename):
        # check if path exist
        if not path.exists(events_filename):
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), events_filename)
        if not path.exists(times_filename):
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), times_filename)
        
        self.events = read_file(events_filename)
        self.times = read_file(times_filename)
        assert self.events.shape == self.times.shape, "Events and Times must be of same dimensions"
    def get_host_count(self):
        if self.events is None or self.times is None:
            return None
        else:
            return len(self.events)

    def plot_data(self):
        plot_host_events(self.events, self.times)

    def describe(self):
        pass


class NTPPData(Dataset):
    def __init__(self, temporal_data, args):
        self.args = args
        self.type = 'train'
        events = temporal_data.events
        times = temporal_data.times
        self.time_step = args['time_step']
        self.host_count = len(events)
        train_interval = int(args['int_count'] * (1 - args['test_size']))
        interval_size = (max([s[-1] for s in times]) + 1) / args['int_count']

        interval_count = np.zeros((self.host_count, args['int_count']),
                                  dtype=int)

        self.train_event, self.train_times = [
            [] for i in range(self.host_count)
        ], [[] for i in range(self.host_count)]
        self.test_event, self.test_times = [[] for i in range(self.host_count)
                                  ], [[] for i in range(self.host_count)]
        for i, host in enumerate(times):
            for j, time_stamp in enumerate(host):
                counter = int(time_stamp / interval_size)
                interval_count[i][counter] += 1
                if counter < train_interval:
                    self.train_times[i].append(times[i][j])
                    # print(type(events[i][j]))
                    self.train_event[i].append(1.0)
                else:
                    self.test_times[i].append(times[i][j])
                    self.test_event[i].append(1.0)
        print(len(self.train_event),len(self.train_event[0]))
        min_count = min([len(x) for x in self.train_event])
        assert min_count >= args[
            'time_step'] + 2, "Time Step should be less than {0}".format(
                min_count - 2)
        self.train_y = compare_interval_count(0, train_interval,
                                              self.host_count, interval_count)

        self.test_y = compare_interval_count(train_interval, args['int_count'],
                                             self.host_count, interval_count)
        self.train_times = [x[:self.time_step + 2] for x in self.train_times]
        self.train_event = [x[:self.time_step + 2] for x in self.train_event]
        

    def getObservation(self):
        return self.train_y, self.test_y

    def startDev(self):
        self.type = 'dev'

    def startTrain(self):
        self.type = 'train'
    
    def change_data(self,times):
        
        # print(self.host_count)
        for i in range(self.host_count):
            last_value = self.train_times[i][-1]
            self.train_times[i].append(last_value+times[i])
        # self.train_times = [x[1:] for x in self.train_times]
            
    def __getitem__(self, idx):
        if self.type == 'train':
            return np.array(self.train_event[idx][:self.time_step + 2]), np.array(
                    self.train_times[idx][-2 - self.time_step:])
        elif self.type =='dev':
            return np.array(self.test_event[idx][:self.time_step + 2]), np.array(
                    self.test_times[idx][:2 + self.time_step])
    def get_times(self):
        return self.train_times
    def __len__(self):
        return self.host_count


if __name__ == '__main__':
    events_filename = "augurnet/data/preprocess/events.txt"
    times_filename  = "augurnet/data/preprocess/times.txt"
    data = TemporalData(events_filename, times_filename)
    print("Host count",data.get_host_count())
    # print(data.events)
        



from distutils.core import setup
setup(
  name = 'augurnet',         
  packages = ['augurnet'],   
  version = '0.2',      
  license='MIT',        
  description = 'Network Traffic Predictions. We develop the product with the mindset of ease. The predictions are done by using the temporal history of the events from the multiple host. The model take into account the effect of multiple host (network) on user’s traffic and contention between different hosts.',   
  author = 'Vedic Partap',                   
  author_email = 'me@vedicpartap.com',      
  url = 'https://github.com/vedic-partap/augurnet',   
  download_url = 'https://github.com/vedic-partap/augurnet/archive/v_02.tar.gz',    
  keywords = ['networ', 'reinforcement-learning', 'predictions', 'generative models'],   
  install_requires=[            
        'numpy>=1.17.2,',
        'torch>=1.3.1',
        'matplotlib>=3.1.1',
        'seaborn>=0.9.0',
        'scikit-learn>=0.21.3'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   
    'Programming Language :: Python :: 3',      
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
